"""Generate code to test the jsonization of classes with descendants."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test the jsonization of classes with descendants."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_jsonization.hpp"
#include "./common_examples.hpp"

#include <{include_prefix_path}/jsonization.hpp>

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace aas = {library_namespace};"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.classes:
        if len(cls.concrete_descendants) == 0:
            continue

        # NOTE (mristin):
        # We can not de-serialize a JSON without a model type. Some classes indeed
        # lack a model type, although they are abstract. This is the case for
        # the classes where no meta-model property uses the abstract class, and only the
        # concrete descendants are used.
        if not cls.serialization.with_model_type:
            continue

        concrete_cls = cls.concrete_descendants[0]

        interface_name = cpp_naming.interface_name(cls.name)
        concrete_interface_name = cpp_naming.interface_name(concrete_cls.name)

        load_min = cpp_naming.function_name(Identifier(f"load_min_{concrete_cls.name}"))

        from_json = cpp_naming.function_name(Identifier(f"{cls.name}_from"))

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test the round-trip of an expected {interface_name}") {{
{I}const std::shared_ptr<
{II}aas::types::{concrete_interface_name}
{I}> concrete_instance(
{II}test::common::examples::{load_min}()
{I});

{I}const nlohmann::json json = aas::jsonization::Serialize(
{II}*concrete_instance
{I});

{I}aas::common::expected<
{II}std::shared_ptr<
{III}aas::types::{interface_name}
{II}>,
{II}aas::jsonization::DeserializationError
{I}> instance = aas::jsonization::{from_json}(
{II}json
{I});

{I}if (!instance.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Failed to deserialize a {interface_name} "
{IIII}"from a minimal {concrete_interface_name}: ",
{IIII}aas::common::WstringToUtf8(instance.error().path.ToWstring()),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(instance.error().cause)
{III})
{II})
{II}CHECK(instance.has_value());
{I}}}

{I}const nlohmann::json another_json = aas::jsonization::Serialize(
{II}**instance
{I});

{I}const std::optional<
{II}std::string
{I}> patch_message = test::common::jsonization::CompareJsons(
{II}json,
{II}another_json
{I});
{I}if (patch_message.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Failed to make a round-trip of a {interface_name} "
{IIII}"over a minimal {concrete_interface_name}: ",
{IIII}*patch_message
{IIII})
{III})
{III}CHECK(!patch_message.has_value());
{I}}}
}}"""
            )
        )

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the JSON de/serialization of concrete classes."""

from typing import List

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the JSON de/serialization of concrete classes.
    """
    blocks = [
        Stripped(
            f"""\
private static void assertSerializeDeserializeEqualsOriginal(JsonNode originalNode, IClass instance, Path path)
{I}throws JsonProcessingException {{
{I}final ObjectMapper objectMapper = new ObjectMapper();

{I}JsonNode serialized = null;
{I}try {{
{II}serialized = Jsonization.Serialize.toJsonObject(instance);
{I}}} catch (Exception exception) {{
{II}fail("Expected no exception upon serialization of an instance " +
{III}"de-serialized from " + path + ", but got: " + exception);
{I}}}

{I}if (serialized == null) {{
{II}fail(
{III}"Unexpected null serialization of an instance from " + path);
{I}}}

assertEquals(objectMapper.readTree(originalNode.toString()), objectMapper.readTree(serialized.toString()));
}}"""
        ),
        Stripped(
            f"""\
private static void assertEqualsExpectedOrRerecordDeserializationException(
{I}Jsonization.DeserializeException exception,
{I}Path path) throws FileNotFoundException, IOException{{
{I}if (exception == null) {{
{II}fail("Expected a Jsonization exception when de-serializing " +
{II}path +
{II}", but got none.");
{I}}} else {{
{II}final Path exceptionPath = Paths.get(path + ".exception");
{II}final String got = exception.getMessage();
{II}if (Common.RECORD_MODE) {{
{III}Files.write(exceptionPath, got.getBytes(StandardCharsets.UTF_8));
{II}}} else {{
if (!Files.exists(exceptionPath)) {{
{I}throw new FileNotFoundException(
{II}"The file with the recorded errors does not exist: "
{III}+ exceptionPath
{III}+ "; maybe you want to set the environment variable "
{III}+ Common.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME);
}}
final String expected =
{I}Files.readAllLines(exceptionPath).stream().collect(Collectors.joining("\\n"));
assertEquals(
{I}expected,
{I}got,
{I}"The expected exception does not match the actual one for the file " + path);
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}Ok() throws IOException {{
{I}final ObjectMapper objectMapper = new ObjectMapper();

{I}final Path searchPath = Paths.get(
{II}Common.TEST_DATA_DIR,
{II}"Json",
{II}"Expected",
{II}{java_common.string_literal(cls_name_json)});
{I}final List<Path> paths = Common.findPaths(searchPath, ".json");

{I}for (Path path : paths) {{
{II}final JsonNode node = objectMapper.readTree(path.toFile());
{II}final {cls_name_java} instance = Jsonization.Deserialize.deserialize{cls_name_java}(node);

{II}final Iterable<Reporting.Error> errorIter = Verification.verify(instance);
{II}final List<Reporting.Error> errors = Common.asList(errorIter);
{II}Common.assertNoVerificationErrors(errors, path);
{I}}}
}} // public void test{cls_name_java}Ok"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}DeserializationFromNonObjectFail() throws IOException {{
{I}final JsonNode node = JsonNodeFactory.instance.textNode("INVALID");

{I}Jsonization.DeserializeException exception = null;
{I}try {{
{II}final {cls_name_java} unused = Jsonization.Deserialize.deserialize{cls_name_java}(node);
{I}}} catch (Jsonization.DeserializeException observedException) {{
{II}exception = observedException;
{I}}}

{I}assert exception != null : "Expected an exception, but got none";
{I}assert exception.getMessage().startsWith("Expected a JsonObject, but got ") :
{II}"Unexpected exception message: " + exception.getMessage();
}} // public void test{cls_name_java}DeserializationFromNonObjectFail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}DeserializationFail() throws IOException {{
{I}for (Path causeDir :
{II}Common.findDirs(
{III}Paths.get(
{IIII}Common.TEST_DATA_DIR,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Unserializable"))) {{
{II}final Path clsDir = causeDir.resolve({java_common.string_literal(cls_name_json)});

{II}if (!Files.exists(clsDir)) {{
{III}// No examples of {cls_name_java} for the failure cause.
{III}continue;
{II}}}

{II}final List<Path> paths = Common.findPaths(clsDir, ".json");
{II}for (Path path : paths) {{
{III}final JsonNode node = CommonJson.readFromFile(path);

{III}Jsonization.DeserializeException exception = null;
{III}try {{
{IIII}final {cls_name_java} var = Jsonization.Deserialize.deserialize{cls_name_java}(node);
{III}}} catch (Jsonization.DeserializeException observedException) {{
{IIII}exception = observedException;
{III}}}

{III}assertEqualsExpectedOrRerecordDeserializationException(
{IIII}exception, path);
{II}}}
{I}}}
}} // public void test{cls_name_java}DeserializationFail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}VerificationFail() throws IOException {{
{I}for (Path causeDir :
{II}Common.findDirs(
{III}Paths.get(
{IIII}Common.TEST_DATA_DIR,
{IIII}"Json",
{IIII}"Unexpected",
{IIII}"Invalid"))) {{
{II}final Path clsDir = causeDir.resolve({java_common.string_literal(cls_name_json)});

{II}if (!Files.exists(clsDir)) {{
{III}// No examples of {cls_name_java} for the failure cause.
{III}continue;
{II}}}

{II}final List<Path> paths = Common.findPaths(clsDir, ".json");
{II}for (Path path : paths) {{
{III}final JsonNode node = CommonJson.readFromFile(path);

{III}final {cls_name_java} instance = Jsonization.Deserialize.deserialize{cls_name_java}(node);

{III}final Iterable<Reporting.Error> errorIter = Verification.verify(instance);
{III}final List<Reporting.Error> errors = Common.asList(errorIter);
{III}Common.assertEqualsExpectedOrRerecordVerificationErrors(errors, path);
{II}}}
{I}}}
}} // public void test{cls_name_java}VerificationFail"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestJsonizationOfConcreteClasses.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.fail;

import {package}.jsonization.Jsonization;
import {package}.reporting.Reporting;
import {package}.types.impl.*;
import {package}.types.model.IClass;
import {package}.verification.Verification;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.JsonNodeFactory;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.stream.Collectors;
import org.junit.jupiter.api.Test;

public class TestJsonizationOfConcreteClasses {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestJsonizationOfConcreteClasses

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``modelType()`` methods."""


import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``modelType()`` methods."""
    blocks = [
        Stripped(
            """\
/**
 * Test `modelType()` methods.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasTypes from "../src/types";
import * as AasStringification from "../src/stringification";
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
    ]  # type: List[Stripped]

    from_string = typescript_naming.function_name(Identifier("model_type_from_string"))

    must_to_string = typescript_naming.function_name(
        Identifier("must_model_type_to_string")
    )

    model_type_enum = typescript_naming.enum_name(Identifier("Model_type"))

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)

        load_minimal_name = typescript_naming.function_name(
            Identifier(f"load_minimal_{concrete_cls.name}")
        )

        model_type_literal = typescript_naming.enum_literal_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
test("model type of {cls_name_typescript}", () => {{
{I}const instance = TestCommonJsonization.{load_minimal_name}();

{I}expect(instance.modelType()).toStrictEqual(
{II}AasTypes.{model_type_enum}.{model_type_literal}
{I});
}});"""
            )
        )

        string_literal = typescript_common.string_literal(
            typescript_naming.enum_literal_name(concrete_cls.name)
        )

        blocks.append(
            Stripped(
                f"""\
test("model type from string of {cls_name_typescript}", () => {{
{I}const text = {string_literal};
{I}const literal = AasStringification.{from_string}(
{II}text
{I});

{I}expect(literal).toStrictEqual(
{II}AasTypes.{model_type_enum}.{model_type_literal}
{I});
}});"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
test("model type to string of {cls_name_typescript}", () => {{
{I}const text = AasStringification.{must_to_string}(
{II}AasTypes.{model_type_enum}.{model_type_literal}
{I});

{I}expect(text).toStrictEqual(
{II}{string_literal}
{I});
}});"""
            )
        )

    blocks.append(
        Stripped(
            f"""\
test("model type from invalid string", () => {{
{I}const text = "This is definitely not a valid model type.";
{I}const literal = AasStringification.{from_string}(
{II}text
{I});

{I}expect(literal).toBeNull();
}});"""
        )
    )

    to_string = typescript_naming.function_name(Identifier("model_type_to_string"))

    blocks.append(
        Stripped(
            f"""\
test("invalid model type to string", () => {{
{I}// The number 9007199254740991 is the maximum safe integer.
{I}const literal = <AasTypes.{model_type_enum}>9007199254740991;
{I}const text = AasStringification.{to_string}(
{II}literal
{I});

{I}expect(text).toBeNull();
}});"""
        )
    )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the XML de/serialization of concrete classes."""

from typing import List

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
    INDENT7 as IIIIIII,
    INDENT8 as IIIIIIII,
    INDENT9 as IIIIIIIII,
    INDENT10 as IIIIIIIIII,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the XML de/serialization of concrete classes.
    """
    blocks = [
        Stripped(
            f"""\
public static Optional<Reporting.Error> checkElementsEqual(
{I}XMLEvent expected, String expectedContent, Map<XMLEvent, String> outputMap) {{
{I}switch (expected.getEventType()) {{
{II}case XMLStreamConstants.START_ELEMENT:
{III}{{
{IIII}final String expectedName = expected.asStartElement().getName().getLocalPart();
{IIII}final Optional<Map.Entry<XMLEvent, String>> got =
{IIIII}outputMap.entrySet().stream()
{IIIIII}.filter(
{IIIIIII}entry ->
{IIIIIIII}entry.getKey().isStartElement()
{IIIIIIIII}&& entry
{IIIIIIIIII}.getKey()
{IIIIIIIIII}.asStartElement()
{IIIIIIIIII}.getName()
{IIIIIIIIII}.getLocalPart()
{IIIIIIIIII}.equals(expectedName))
{IIIIII}.filter(entry -> entry.getValue().equals(expectedContent))
{IIIIIII}.findAny();
{IIII}if (!got.isPresent()) {{
{IIIII}final Reporting.Error error =
{IIIIII}new Reporting.Error(
{IIIIIII}"Missing start element "
{IIIIIIII}+ expectedName
{IIIIIIII}+ " in with content: "
{IIIIIIII}+ expectedContent);
{IIIII}return Optional.of(error);
{IIII}}}
{IIII}outputMap.remove(got.get().getKey());
{IIII}return Optional.empty();
{III}}}
{II}case XMLStreamConstants.END_ELEMENT:
{III}{{
{IIII}final String expectedName = expected.asEndElement().getName().getLocalPart();
{IIII}final Optional<Map.Entry<XMLEvent, String>> got =
{IIIII}outputMap.entrySet().stream()
{IIIIII}.filter(
{IIIIIII}entry ->
{IIIIIIII}entry.getKey().isEndElement()
{IIIIIIIII}&& entry
{IIIIIIIIII}.getKey()
{IIIIIIIIII}.asEndElement()
{IIIIIIIIII}.getName()
{IIIIIIIIII}.getLocalPart()
{IIIIIIIIII}.equals(expectedName))
{IIIIII}.findAny();
{IIII}if (!got.isPresent()) {{
{IIIII}final Reporting.Error error =
{IIIIII}new Reporting.Error("Missing end element " + expectedName);
{IIIII}return Optional.of(error);
{IIII}}}
{IIII}outputMap.remove(got.get().getKey());
{IIII}return Optional.empty();
{III}}}
{II}default:
{III}{{
{IIII}throw new IllegalStateException("Unexpected event type in check elements equal.");
{III}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private static String readContent(XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();
{I}while (reader.hasNext() && reader.peek().isCharacters()
{III}&& !reader.peek().asCharacters().isWhiteSpace()
{III}|| reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{

{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}
{I}return content.toString();
}}"""
        ),
        Stripped(
            f"""\
private static Map<XMLEvent, String> buildElementsMap(XMLEventReader reader) throws XMLStreamException {{
{I}final Map<XMLEvent, String> result = new LinkedHashMap<>();
{I}while (reader.hasNext()) {{
{II}final XMLEvent current = reader.nextEvent();
{II}if (current.isStartElement()) {{
{III}result.put(current, readContent(reader));
{II}}} else if (current.isEndElement()) {{
{III}result.put(current, "");
{II}}}
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
private static void assertSerializeDeserializeEqualsOriginal(IClass instance, Path path)
{I}throws XMLStreamException, IOException {{
{I}// Serialize
{I}final StringWriter stringOut = new StringWriter();
{I}final XMLOutputFactory outputFactory = XMLOutputFactory.newFactory();
{I}final XMLStreamWriter xmlStreamWriter = outputFactory.createXMLStreamWriter(stringOut);

{I}Xmlization.Serialize.to(instance, xmlStreamWriter);

{I}final String outputText = stringOut.toString();

// Compare expected == output
final XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
final XMLEventReader outputReader =
{I}xmlInputFactory.createXMLEventReader(new StringReader(outputText));
final Map<XMLEvent, String> outputMap = buildElementsMap(outputReader);

// check output for aas-name-space
for (XMLEvent event : outputMap.keySet()) {{
{I}if (event.isStartElement()) {{
{II}assertEquals(Xmlization.AAS_NAME_SPACE, event.asStartElement().getName().getNamespaceURI());
{I}}}
{I}if (event.isEndElement()) {{
{II}assertEquals(Xmlization.AAS_NAME_SPACE, event.asEndElement().getName().getNamespaceURI());
{I}}}
}}

final XMLEventReader expectedReader =
{I}xmlInputFactory.createXMLEventReader(Files.newInputStream(path));
final Map<XMLEvent, String> expectedMap = buildElementsMap(expectedReader);

if (expectedMap.size() != outputMap.size()) {{
{I}fail(
{II}"Mismatch in element size expected "
{III}+ expectedMap.size()
{III}+ " but got "
{III}+ outputMap.size());
}}

expectedMap.forEach(
{I}(xmlEvent, content) -> {{
{II}final Optional<Reporting.Error> inequalityError =
{IIII}checkElementsEqual(xmlEvent, content, outputMap);
{II}inequalityError.ifPresent(
{III}error ->
{IIII}fail(
{IIIII}"The original XML from "
{IIIIII} + path
{IIIIII} + " is unequal the serialized XML: "
{IIIIII} + error.getCause()));
{I}}});
}}"""
        ),
        Stripped(
            f"""\
private static void assertEqualsExpectedOrRerecordDeserializationException(
{I}Xmlization.DeserializeException exception,
{I}Path path) throws IOException {{
{I}if (exception == null) {{
{II}fail("Expected a Xmlization exception when de-serializing " + path + ", but got none.");
{I}}} else {{
{II}final Path exceptionPath = Paths.get(path + ".exception");
{II}final String got = exception.getMessage();
{II}if (Common.RECORD_MODE) {{
{III}Files.write(exceptionPath, got.getBytes(StandardCharsets.UTF_8));
{II}}} else {{
{III}if (!Files.exists(exceptionPath)) {{
{IIII}throw new FileNotFoundException(
{IIIII}"The file with the recorded exception does not exist: "
{IIIII}+ exceptionPath
{IIIII}+ "; maybe you want to set the environment variable"
{IIIII}+ Common.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME
{IIIII}+ "?");
{III}}}

{III}final String expected = String.join("\\n", Files.readAllLines(exceptionPath));
{III}assertEquals(
{IIII}expected.replace("\\n", ""),
{IIII}got.replace("\\n", ""),
{IIII}"The expected exception does not match the actual one for the file " + path);
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)
        cls_name_xml = naming.xml_class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}Ok() throws IOException, XMLStreamException {{
{I}final Path searchPath =
{II}Paths.get(Common.TEST_DATA_DIR, "Xml", "Expected", {java_common.string_literal(cls_name_xml)});
{I}final List<Path> paths = Common.findPaths(searchPath, ".xml");

{I}for (Path path : paths) {{
{II}final XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
{II}final XMLEventReader xmlReader =
{III}xmlInputFactory.createXMLEventReader(Files.newInputStream(path));

{II}final {cls_name_java} instance =
{III}Xmlization.Deserialize.deserialize{cls_name_java}(xmlReader);

{II}final Iterable<Reporting.Error> errorIter = Verification.verify(instance);
{II}final List<Reporting.Error> errors = Common.asList(errorIter);
{II}Common.assertNoVerificationErrors(errors, path);

{II}assertSerializeDeserializeEqualsOriginal(instance, path);
{I}}}
}} // public void test{cls_name_java}Ok"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}DeserializationFail() throws IOException, XMLStreamException {{
{I}for (
{II}Path causeDir :
{II}Common.findDirs(
{III}Paths.get(
{IIII}Common.TEST_DATA_DIR,
{IIII}"Xml",
{IIII}"Unexpected",
{IIII}"Unserializable"))) {{
{II}final Path clsDir =
{III}causeDir.resolve({java_common.string_literal(cls_name_xml)});

{II}if (!Files.exists(clsDir)) {{
{III}// No examples of {cls_name_java} for the failure cause.
{III}continue;
{II}}}

{II}final List<Path> paths = Common.findPaths(clsDir, ".xml");
{II}for (Path path : paths) {{
{III}final XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
{III}final XMLEventReader xmlReader =
{IIII}xmlInputFactory.createXMLEventReader(Files.newInputStream(path));

{III}Xmlization.DeserializeException exception = null;

{III}try {{
{IIII}Xmlization.Deserialize.deserialize{cls_name_java}(xmlReader);
{III}}} catch (Xmlization.DeserializeException observedException) {{
{IIII}exception = observedException;
{III}}}

{III}assertEqualsExpectedOrRerecordDeserializationException(exception, path);
{II}}}
{I}}}
}}  // public void test{cls_name_java}DeserializationFail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name_java}VerificationFail() throws IOException, XMLStreamException {{
{I}for (
{II}Path causeDir :
{II}Common.findDirs(
{III}Paths.get(
{IIII}Common.TEST_DATA_DIR,
{IIII}"Xml",
{IIII}"Unexpected",
{IIII}"Invalid"))) {{
{II}final Path clsDir = causeDir.resolve(
{III}{java_common.string_literal(cls_name_xml)});

{II}if (!Files.exists(clsDir)) {{
{III}// No examples of {cls_name_java} for the failure cause.
{III}continue;
{II}}}

{II}final List<Path> paths = Common.findPaths(clsDir, ".xml");
{II}for (Path path : paths) {{
{III}final XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
{III}final XMLEventReader xmlReader =
{IIII}xmlInputFactory.createXMLEventReader(Files.newInputStream(path));

{III}final {cls_name_java} instance =
{IIII}Xmlization.Deserialize.deserialize{cls_name_java}(xmlReader);

{III}final Iterable<Reporting.Error> errorIter = Verification.verify(instance);
{III}final List<Reporting.Error> errors = Common.asList(errorIter);
{III}Common.assertEqualsExpectedOrRerecordVerificationErrors(errors, path);
{II}}}
{I}}}
}} // public void test{cls_name_java}VerificationFail"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestXmlizationOfConcreteClasses.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.fail;

import {package}.reporting.Reporting;
import {package}.types.impl.*;
import {package}.types.model.IClass;
import {package}.verification.Verification;
import {package}.xmlization.Xmlization;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import javax.xml.stream.*;
import javax.xml.stream.events.XMLEvent;
import org.junit.jupiter.api.Test;

public class TestXmlizationOfConcreteClasses {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestXmlizationOfConcreteClasses

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code for common functionality shared across the tests."""

import io
from typing import List, Optional, Tuple

from icontract import ensure

from aas_core_codegen import specific_implementations
from aas_core_codegen.common import Stripped, Error
from aas_core_codegen.typescript import common as typescript_common
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


# fmt: off
@ensure(
    lambda result: not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def generate(
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate code for common functionality shared across the tests."""
    package_identifier_key = specific_implementations.ImplementationKey(
        "package_identifier.txt"
    )

    package_identifier = spec_impls.get(package_identifier_key, None)
    if package_identifier is None:
        return None, [
            Error(
                None,
                f"The package identifier snippet is missing "
                f"in the specific implementations: {package_identifier_key}",
            )
        ]

    assert package_identifier is not None

    environment_variable_prefix = typescript_common.environment_variable_prefix(
        package_identifier
    )

    blocks = [
        Stripped(
            """\
/**
 * Provide common functionality to be re-used across different tests
 * such as reading of commonly-used environment variables.
 */"""
        ),
        Stripped(
            """\
import * as fs from "fs";
import * as path from "path";"""
        ),
        Stripped(
            """\
import * as AasTypes from "../src/types";
import * as AasVerification from "../src/verification";
import * as AasJsonization from "../src/jsonization";
import { Path } from "../src/jsonization";"""
        ),
        Stripped(
            f"""\
// NOTE (mristin):
// It is tedious to record manually all the expected error messages. Therefore we include this variable
// to steer the automatic recording. We intentionally inter-twine the recording code with the test code
// to keep them close to each other so that they are easier to maintain.
export const RECORD_MODE_ENVIRONMENT_VARIABLE_NAME =
{I}"{environment_variable_prefix}_TEST_RECORD_MODE";"""
        ),
        Stripped(
            f"""\
const RECORD_MODE_TEXT =
{I}process.env[RECORD_MODE_ENVIRONMENT_VARIABLE_NAME]?.toLowerCase();
export const RECORD_MODE: boolean =
{I}RECORD_MODE_TEXT === "true" || RECORD_MODE_TEXT === "1" || RECORD_MODE_TEXT === "on";"""
        ),
        Stripped(
            f"""\
export const TEST_DATA_DIR = process.env[
{I}"{environment_variable_prefix}_TEST_DATA_DIR"
];
if (TEST_DATA_DIR === null || TEST_DATA_DIR === undefined) {{
{I}throw new Error(
{II}"The path to the test data directory is missing in the environment: " +
{III}"AAS_CORE3_0_TYPESCRIPT_TEST_DATA_DIR"
{I});
}}
if (!fs.existsSync(TEST_DATA_DIR)) {{
{I}throw new Error(
{II}"The path read from environment variable " +
{III}"AAS_CORE3_0_TYPESCRIPT_TEST_DATA_DIR does not exist: " +
{III}TEST_DATA_DIR
{I});
}}
if (!fs.lstatSync(TEST_DATA_DIR).isDirectory()) {{
{I}throw new Error(
{II}"The path read from environment variable " +
{III}"AAS_CORE3_0_TYPESCRIPT_TEST_DATA_DIR is not a directory: " +
{III}TEST_DATA_DIR
{I});
}}"""
        ),
        Stripped(
            f"""\
/**
 * Read a JSON value from a file.
 *
 * @param aPath - to the file
 * @returns a JSON value read from the file
 */
export function readJsonFromFileSync(aPath: string): AasJsonization.JsonValue {{
{I}const text = fs.readFileSync(aPath, "utf-8");
{I}let jsonable: AasJsonization.JsonValue | null = null;
{I}try {{
{II}jsonable = JSON.parse(text);
{I}}} catch (error) {{
{II}throw new Error(`Failed to parse JSON from: ${{aPath}}`);
{I}}}

{I}if (jsonable === null) {{
{II}throw new Error(`Unexpected null value as JSON from: ${{aPath}}`);
{I}}}

{I}return <AasJsonization.JsonValue>jsonable;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Find the first instance beneath and including the `container` which satisfies
 * `condition`.
 *
 * @param container - where to search
 * @param condition - that needs to be fulfilled
 * @throws an {{@link Error}} if no instance could be found which satisfies
 * `condition`
 */
export function mustFind(
{I}container: AasTypes.Class,
{I}condition: (instance: AasTypes.Class) => boolean
) {{
{I}if (condition(container)) {{
{II}return container;
{I}}}

{I}for (const instance of container.descend()) {{
{II}if (condition(instance)) {{
{III}return instance;
{II}}}
{I}}}

{I}throw new Error("No instance could be found which satisfies the condition.");
}}"""
        ),
        Stripped(
            f"""\
/**
 * Assert that there are no verification errors in the `iterable`.
 *
 * @param errors - iterable of verification errors
 * @param aPath - to the file specifying the instance
 * @throws an {{@link Error}} with an informative message
 */
export function assertNoVerificationErrors(
{I}errors: IterableIterator<AasVerification.VerificationError>,
{I}aPath: string
): void {{
{I}const errorArray = errors instanceof Array ? errors : Array.from(errors);

{I}if (errorArray.length !== 0) {{
{II}let message =
{III}"Expected no errors when verifying the instance de-serialized " +
{III}`from ${{aPath}}, but got ${{errorArray.length}} error(s):`;

{II}for (let i = 0; i < errorArray.length; i++) {{
{III}const error = errorArray[i];

{III}message += `\\n\\nError ${{i + 1}}:\\n${{error.path}}: ${{error.message}}`;
{II}}}
{II}throw new Error(message);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Assert that `errors` either correspond to the errors recorded to the disk,
 * or re-record the errors, if {{@link RECORD_MODE}} is set.
 *
 * @param errors - iterable of verification errors on an instance
 * @param aPath - to the instance
 * @throws an {{@link Error}} if {{@link RECORD_MODE}} unset and the observed and
 * recorded errors do not coincide
 */
export function assertExpectedOrRecordedVerificationErrors(
{I}errors: IterableIterator<AasVerification.VerificationError>,
{I}aPath: string
): void {{
{I}const errorArray = errors instanceof Array ? errors : Array.from(errors);

{I}if (errorArray.length === 0) {{
{II}throw new Error(
{III}"Expected at least one verification error when " +
{IIII}`verifying ${{path}}, but got none`
{II});
{I}}}

{I}const got = errorArray.map((error) => `${{error.path}}: ${{error.message}}`).join(";\\n");

{I}const errorsPath = aPath + ".errors";
{I}if (RECORD_MODE) {{
{II}fs.writeFileSync(errorsPath, got, "utf-8");
{I}}} else {{
{II}if (!fs.existsSync(errorsPath)) {{
{III}throw new Error(
{IIII}`The file with the recorded verification errors does not ` +
{IIIII}`exist: ${{errorsPath}}; you probably want to set the environment ` +
{IIIII}`variable ${{RECORD_MODE_ENVIRONMENT_VARIABLE_NAME}}?`
{III});
{II}}}

{II}const expected = fs.readFileSync(errorsPath, "utf-8");
{II}if (expected !== got) {{
{III}throw new Error(
{IIII}`Expected verification errors from ${{path}}:\\n` +
{IIIII}`${{expected}}\\n, but got:\\n${{got}}`
{III});
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent the `instance` as a human-readable line of an iteration trace.
 *
 * @param instance - to leave a mark in the trace
 * @returns the mark in the trace
 */
export function traceMark(instance: AasTypes.Class): string {{
{I}return instance.constructor.name;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Iterate over all the files beneath `directory` with the given `suffix` path.
 *
 * @param directory - to iterate through recursively
 * @param suffix - expected suffix of the file name
 */
export function* findFilesBySuffixRecursively(
{I}directory: string,
{I}suffix: string
): IterableIterator<string> {{
{I}for (const filename of fs.readdirSync(directory)) {{
{II}const pth = path.join(directory, filename);
{II}const stat = fs.lstatSync(pth);
{II}if (stat.isDirectory()) {{
{III}yield* findFilesBySuffixRecursively(pth, suffix);
{II}}} else {{
{III}if (filename.endsWith(suffix)) {{
{IIII}yield pth;
{III}}}
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Iterate over all the immediate subdirectories `directory`.
 *
 * @param directory - to iterate through
 */
export function* findImmediateSubdirectories(
{I}directory: string
): IterableIterator<string> {{
{I}for (const filename of fs.readdirSync(directory)) {{
{II}const pth = path.join(directory, filename);
{II}const stat = fs.lstatSync(pth);
{II}if (stat.isDirectory()) {{
{III}yield pth;
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Signal that two JSON-able structures are unequal.
 */
export class InequalityError {{
{I}/**
{I} * Human-readable explanation of the error
{I} */
{I}readonly message: string;

{I}/**
{I} * Relative path to the erroneous value
{I} */
{I}readonly path: AasJsonization.Path;

{I}constructor(message: string, path: AasJsonization.Path | null = null) {{
{II}this.message = message;
{II}this.path = path ?? new Path();
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Check that the `expected` JSON-able structure strictly equals `got`
 * JSON-able structure.
 *
 * @param expected - JSON-able structure
 * @param got - JSON-able structure
 */
export function checkJsonablesEqual(
{I}expected: AasJsonization.JsonValue | null,
{I}got: AasJsonization.JsonValue | null
): AasJsonization.DeserializationError | null {{
{I}if (
{II}expected === null ||
{II}typeof expected === "boolean" ||
{II}typeof expected === "number" ||
{II}typeof expected === "string"
{I}) {{
{II}if (expected !== got) {{
{III}if (
{IIII}got === null ||
{IIII}typeof got === "boolean" ||
{IIII}typeof got === "number" ||
{IIII}typeof got === "string"
{III}) {{
{IIII}return new InequalityError(
{IIIII}`Expected ${{JSON.stringify(expected)}}, ` + `but got ${{JSON.stringify(got)}}`
{IIII});
{III}}} else {{
{IIII}return new InequalityError(
{IIIII}`Expected ${{JSON.stringify(expected)}}, ` +
{IIIIII}`but got an instance of ${{got.constructor.name}}`
{IIII});
{III}}}
{II}}}
{I}}} else if (
{II}typeof expected === "object" &&
{II}typeof expected[Symbol.iterator] === "function"
{I}) {{
{II}if (typeof got !== "object") {{
{III}return new InequalityError(
{IIII}`Expected an iterable, ` + `but got ${{JSON.stringify(got)}}`
{III});
{II}}}

{II}if (typeof got[Symbol.iterator] !== "function") {{
{III}return new InequalityError(
{IIII}`Expected an iterable, ` + `but got an instance of ${{got.constructor.name}}`
{III});
{II}}}

{II}const expectedIt = <Iterator<AasJsonization.JsonValue>>expected[Symbol.iterator]();

{II}const gotIt = <Iterator<AasJsonization.JsonValue>>got[Symbol.iterator]();

{II}let i = 0;
{II}// eslint-disable-next-line no-constant-condition
{II}while (true) {{
{III}const expectedResult = expectedIt.next();
{III}const gotResult = gotIt.next();

{III}if (expectedResult.done && gotResult.done) {{
{IIII}break;
{III}}}

{III}if (expectedResult.done && !gotResult.done) {{
{IIII}return new InequalityError(
{IIIII}`Expected an iterable with ${{i + 1}} items, ` +
{IIIIII}`but got an iterable with more items`
{IIII});
{III}}}

{III}if (!expectedResult.done && gotResult.done) {{
{IIII}return new InequalityError(
{IIIII}`Expected an iterable with more than ${{i + 1}} item(s), ` +
{IIIIII}`but got an iterable with only ${{i + 1}} item(s)`
{IIII});
{III}}}

{III}const expectedItem = expectedResult.value;
{III}const gotItem = expectedResult.value;

{III}const error = checkJsonablesEqual(expectedItem, gotItem);
{III}if (error !== null) {{
{IIII}error.path.prepend(
{IIIII}new AasJsonization.IndexSegment(<AasJsonization.JsonArray>expected, i)
{IIII});
{IIII}return error;
{III}}}

{III}i++;
{II}}}
{I}}} else if (typeof expected === "object") {{
{II}const expectedKeys = new Set<string>(
{III}Object.keys(expected).filter((key) =>
{IIII}Object.prototype.hasOwnProperty.call(expected, key)
{III})
{II});

{II}const gotKeys = new Set<string>(
{III}Object.keys(expected).filter((key) =>
{IIII}Object.prototype.hasOwnProperty.call(got, key)
{III})
{II});

{II}for (const key of expectedKeys) {{
{III}if (!gotKeys.has(key)) {{
{IIII}return new InequalityError(
{IIIII}`Expected an object with key ${{JSON.stringify(key)}}, ` +
{IIIIII}`but got an object without that key`
{IIII});
{III}}}
{II}}}

{II}for (const key of gotKeys) {{
{III}if (!expectedKeys.has(key)) {{
{IIII}return new InequalityError(
{IIIII}`Expected an object without the key ${{JSON.stringify(key)}}, ` +
{IIIIII}`but got an object with that key`
{IIII});
{III}}}
{II}}}

{II}for (const key of expectedKeys) {{
{III}const expectedValue = expected[key];
{III}const gotValue = got[key];

{III}const error = checkJsonablesEqual(expectedValue, gotValue);
{III}if (error !== null) {{
{IIII}error.path.prepend(
{IIIII}new AasJsonization.PropertySegment(<AasJsonization.JsonObject>expected, key)
{IIII});
{IIII}return error;
{III}}}
{II}}}
{I}}} else {{
{II}throw new Error(`Unexpected expected value: ${{expected}}`);
{I}}}

{I}return null;
}}"""
        ),
    ]  # type: List[Stripped]
    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code for common XML de/serialization shared across the tests."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
    INDENT7 as IIIIIII,
    INDENT8 as IIIIIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(library_namespace: Stripped) -> str:
    """Generate header for common XML de/serialization shared across the tests."""
    include_guard_var = cpp_common.include_guard_var(
        Stripped("test::common::xmlization")
    )

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
/**
 * Provide methods which are used throughout the jsonization tests.
 */
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        Stripped(
            f"""\
#include <{include_prefix_path}/types.hpp>"""
        ),
        Stripped(
            """\
#include <filesystem>
#include <string>
#include <vector>"""
        ),
        Stripped(
            """\
namespace test {
namespace common {
namespace xmlization {"""
        ),
        Stripped(
            f"""\
/**
 * Read the content of the `path` and parse it as XML representation of an instance.
 */
std::shared_ptr<
{I}{library_namespace}::types::IClass
> MustReadInstance(
{I}const std::filesystem::path& path
);"""
        ),
        Stripped(
            f"""\
/**
 * Canonicalize the `xml` by stripping away any XML text between the stop and start
 * nodes.
 */
std::string CanonicalizeXml(
{I}const std::string& xml
);"""
        ),
        Stripped(
            f"""\
std::shared_ptr<
{I}{library_namespace}::types::IClass
> MustDeserializeFile(
{I}const std::filesystem::path& path
);"""
        ),
        Stripped(
            """\
}  // namespace xmlization
}  // namespace common
}  // namespace test"""
        ),
        Stripped(
            f"""\
#endif  // {include_guard_var}"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate implementation for common XML de/serialization shared across the tests."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            '''\
#include "./common_xmlization.hpp"
#include "./common.hpp"'''
        ),
        Stripped(
            f"""\
#include <{include_prefix_path}/xmlization.hpp>"""
        ),
        Stripped(
            """\
#pragma warning(push, 0)
#include <expat.h>
#pragma warning(pop)"""
        ),
        Stripped(
            """\
#include <fstream>"""
        ),
        Stripped(
            f"""\
namespace aas = {library_namespace};"""
        ),
        Stripped(
            """\
namespace test {
namespace common {
namespace xmlization {"""
        ),
        Stripped(
            f"""\
std::shared_ptr<
{I}{library_namespace}::types::IClass
> MustReadInstance(
{I}const std::filesystem::path& path
) {{
{I}std::ifstream ifs(path);

{I}aas::common::expected<
{II}std::shared_ptr<aas::types::IClass>,
{II}aas::xmlization::DeserializationError
{I}> instance = aas::xmlization::From(
{II}ifs
{I});

{I}if (ifs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read XML from ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (!instance.has_value()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to parse the instance from the XML file ",
{IIII}path.string(),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(instance.error().path.ToWstring()),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(instance.error().cause)
{III})
{II});
{I}}}

{I}return *instance;
}}"""
        ),
        Stripped(
            """\
namespace canonicalizer {"""
        ),
        Stripped(
            f"""\
struct INode {{
{I}std::string text;

{I}INode(
{II}std::string a_text
{I}) :
{II}text(a_text) {{
{II}// Intentionally empty.
{I}}}

{I}virtual std::unique_ptr<INode> Copy() const = 0;
{I}virtual ~INode() = default;
}};"""
        ),
        Stripped(
            f"""\
struct StartNode : INode {{
{I}std::string name;

{I}StartNode(
{II}std::string a_name,
{II}std::string a_text
{I}) :
{II}INode(std::move(a_text)),
{II}name(std::move(a_name)) {{
{II}// Intentionally empty.
{I}}}

{I}std::unique_ptr<INode> Copy() const override {{
{II}return std::make_unique<StartNode>(*this);
{I}}}

{I}~StartNode() override = default;
}};"""
        ),
        Stripped(
            f"""\
struct TextNode : INode {{
{I}TextNode(
{II}std::string a_text
{I}) : INode(std::move(a_text)) {{
{II}// Intentionally empty.
{I}}}

{I}std::unique_ptr<INode> Copy() const override {{
{II}return std::make_unique<TextNode>(*this);
{I}}}

{I}~TextNode() override = default;
}};"""
        ),
        Stripped(
            f"""\
struct StopNode : INode {{
{I}std::string name;

{I}StopNode(
{II}std::string a_name,
{II}std::string a_text
{I}) :
{II}INode(std::move(a_text)),
{II}name(std::move(a_name)) {{
{II}// Intentionally empty.
{I}}}

{I}std::unique_ptr<INode> Copy() const override {{
{II}return std::make_unique<StopNode>(*this);
{I}}}

{I}~StopNode() override = default;
}};"""
        ),
        Stripped(
            f"""\
/**
 * Structure the data passed over to Expat XML reader.
 */
struct OurData {{
{I}XML_Parser parser;
{I}std::optional<std::string> error;

{I}std::deque<std::unique_ptr<INode> > nodes;

{I}OurData(
{II}XML_Parser a_parser
{I}) :
{II}parser(a_parser) {{
{II}// Intentionally empty.
{I}}}
}};  // struct OurData"""
        ),
        Stripped(
            f"""\
aas::common::expected<
{I}std::string,
{I}std::string
> EscapeForXmlAttribute(const std::string_view& text) {{
{I}// NOTE (mristin):
{I}// See: https://stackoverflow.com/questions/19766669/which-characters-are-permitted-in-xml-attributes
{I}//
{I}// In addition, we assume that we will always surround an attribute by a double-quote.

{I}size_t size = 0;
{I}for (char character : text
{II}) {{
{II}switch (character) {{
{III}case '<':size += sizeof("&lt;");
{IIII}break;
{III}case '"':size += sizeof("&quot;");
{IIII}break;
{III}case '\\0':
{IIII}return aas::common::unexpected<std::string>(
{IIIII}"Unexpected zero character in the text to be escaped "
{IIIII}"for XML attribute. The XML format does not allow zero characters, "
{IIIII}"not even in escaped form."
{IIII});
{III}default:++size;
{II}}}
{I}}}

{I}std::string result;
{I}result.reserve(size);

{I}for (char character : text
{II}) {{
{II}switch (character) {{
{III}case '<':result.append("&lt;");
{IIII}break;
{III}case '"':result.append("&quot;");
{IIII}break;
{III}default:result.push_back(character);
{II}}}
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
aas::common::expected<
{I}std::string,
{I}std::string
> EscapeForXmlText(const std::string_view& text) {{
{I}size_t size = 0;
{I}for (char character : text
{II}) {{
{II}switch (character) {{
{III}case '<':size += sizeof("&lt;");
{IIII}break;
{III}case '>':size += sizeof("&gt;");
{IIII}break;
{III}case '&':size += sizeof("&amp;");
{IIII}break;
{III}case '\\'':size += sizeof("&apos;");
{IIII}break;
{III}case '"':size += sizeof("&quot;");
{IIII}break;
{III}case '\\0':
{IIII}return aas::common::unexpected(
{IIIII}"Unexpected zero character in the text to be escaped "
{IIIII}"for XML attribute. The XML format does not allow zero characters, "
{IIIII}"not even in escaped form."
{IIII});
{III}default:++size;
{II}}}
{I}}}

{I}if (text.size() == size) {{
{II}return std::string(text);
{I}}}

{I}std::string result;
{I}result.reserve(size);

{I}for (char character : text
{II}) {{
{II}switch (character) {{
{III}case '<':result.append("&lt;");
{IIII}break;
{III}case '>':result.append("&gt;");
{IIII}break;
{III}case '&':result.append("&amp;");
{IIII}break;
{III}case '\\'':result.append("&apos;");
{IIII}break;
{III}case '"':result.append("&quot;");
{IIII}break;
{III}default:result.push_back(character);
{II}}}
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
void XMLCALL OnStartElement(
{I}void* user_data,
{I}const char* name,
{I}const char* attributes[]
) {{
{I}auto our_data = static_cast<OurData*>(user_data);

{I}// NOTE (mristin):
{I}// Since Expat continues parsing and adding nodes even if the parsing is
{I}// suspended (see the documentation of `XML_StopParser`), we have to ignore
{I}// any further events.
{I}if (our_data->error.has_value()) {{
{II}return;
{I}}}

{I}std::stringstream ss;
{I}ss << "<" << name;

{I}for (size_t i = 0; attributes[i] != nullptr; i += 2
{II}) {{
{II}const char* attribute_name = attributes[i];
{II}const char* attribute_value = attributes[i + 1];

{II}aas::common::expected<
{III}std::string,
{III}std::string
{II}> escaped = EscapeForXmlAttribute(
{III}std::string_view(
{IIII}attribute_value,
{IIII}strlen(attribute_value)
{III})
{II});
{II}if (!escaped.has_value()) {{
{III}our_data->error = escaped.error();
{III}XML_StopParser(our_data->parser, false);
{III}return;
{II}}}

{II}ss
{III}<< " "
{III}<< attribute_name << "=\\""
{III}<< *escaped
{III}<< "\\"";
{I}}}

{I}ss << ">";

{I}our_data->nodes.emplace_back(
{II}aas::common::make_unique<StartNode>(
{III}name,
{III}std::move(ss.str())
{II})
{I});
}}"""
        ),
        Stripped(
            f"""\
void XMLCALL OnStopElement(
{I}void* user_data,
{I}const char* name
) {{
{I}auto* our_data = static_cast<OurData*>(user_data);

{I}// NOTE (mristin):
{I}// Since Expat continues parsing and adding nodes even if the parsing is
{I}// suspended (see the documentation of `XML_StopParser`), we have to ignore
{I}// any further events.
{I}if (our_data->error.has_value()) {{
{II}return;
{I}}}

{I}our_data->nodes.emplace_back(
{II}aas::common::make_unique<StopNode>(
{III}name,
{III}aas::common::Concat(
{IIII}"</", name, ">"
{III})
{II})
{I});
}}"""
        ),
        Stripped(
            f"""\
void XMLCALL OnText(
{I}void* user_data,
{I}const char* val,
{I}int len
) {{
{I}auto our_data = static_cast<OurData*>(user_data);

{I}// NOTE (mristin):
{I}// Since Expat continues parsing and adding nodes even if the parsing is
{I}// suspended (see the documentation of `XML_StopParser`), we have to ignore
{I}// any further events.
{I}if (our_data->error.has_value()) {{
{II}return;
{I}}}

{I}aas::common::expected<
{II}std::string,
{II}std::string
{I}> escaped = EscapeForXmlText(std::string_view(val, len));

{I}if (!escaped.has_value()) {{
{II}our_data->error = std::move(escaped.error());
{II}XML_StopParser(our_data->parser, false);
{II}return;
{I}}}

{I}our_data->nodes.emplace_back(
{II}aas::common::make_unique<TextNode>(
{III}std::move(*escaped)
{II})
{I});
}}"""
        ),
        Stripped(
            f"""\
std::deque<std::unique_ptr<INode> > MergeTextNodes(
{I}const std::deque<std::unique_ptr<INode> >& nodes
) {{
{I}std::deque<std::string> parts;

{I}std::deque<std::unique_ptr<INode> > result;

{I}for (const std::unique_ptr<INode>& node : nodes
{II}) {{
{II}auto start_node = dynamic_cast<const StartNode*>(node.get());
{II}auto stop_node = dynamic_cast<const StopNode*>(node.get());

{II}if (start_node != nullptr || stop_node != nullptr) {{
{III}if (!parts.empty()) {{
{IIII}result.emplace_back(
{IIIII}std::make_unique<TextNode>(
{IIIIII}test::common::JoinStrings(parts, "")
{IIIII})
{IIII});
{IIII}parts.clear();
{III}}}

{III}result.emplace_back(node->Copy());

{III}continue;
{II}}}

{II}auto text_node = dynamic_cast<const TextNode*>(node.get());
{II}if (text_node != nullptr) {{
{III}parts.push_back(text_node->text);
{II}}}
{I}}}

{I}if (!parts.empty()) {{
{II}result.emplace_back(
{III}std::make_unique<TextNode>(
{IIII}test::common::JoinStrings(parts, "")
{III})
{II});
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
bool IsWhitespace(const std::string& text) {{
{I}for (const char character : text
{II}) {{
{II}switch (character) {{
{III}case '\\n':continue;
{III}case '\\t':continue;
{III}case '\\r':continue;
{III}case ' ':continue;
{III}default:return false;
{II}}}
{I}}}
{I}return true;
}}"""
        ),
        Stripped(
            f"""\
/**
 * Remove the whitespace between two start/stop nodes with differing names.
 */
std::deque<std::unique_ptr<INode> > RemoveNonSemanticWhiteSpace(
{I}const std::deque<std::unique_ptr<INode> >& nodes
) {{
{I}std::deque<std::unique_ptr<INode> > result;

{I}if (nodes.size() <= 2) {{
{II}for (const auto& node : nodes
{III}) {{
{III}result.emplace_back(node->Copy());
{II}}}
{II}return result;
{I}}}

{I}INode* previous = nullptr;
{I}StartNode* previous_as_start = nullptr;

{I}auto it = nodes.begin();
{I}INode* current = it->get();
{I}StartNode* current_as_start = dynamic_cast<StartNode*>(current);
{I}TextNode* current_as_text = dynamic_cast<TextNode*>(current);

{I}++it;
{I}INode* lookahead = it->get();
{I}StartNode* lookahead_as_start = dynamic_cast<StartNode*>(lookahead);
{I}StopNode* lookahead_as_stop = dynamic_cast<StopNode*>(lookahead);
{I}TextNode* lookahead_as_text = dynamic_cast<TextNode*>(lookahead);

{I}++it;

{I}while (true) {{
{II}// region Determine what to do

{II}// If not set, we include the current node in the result.
{II}bool ignore = (
{III}current_as_text != nullptr
{IIII}&& IsWhitespace(current_as_text->text)
{IIII}&& (
{IIIII}previous == nullptr
{IIIIII}|| lookahead == nullptr
{IIIIII}|| !(
{IIIIIII}previous_as_start != nullptr
{IIIIIIII}&& lookahead_as_stop != nullptr
{IIIIIIII}&& previous_as_start->name == lookahead_as_stop->name
{IIIIII})
{IIII})
{II});
{II}// endregion

{II}// region Act
{II}if (!ignore) {{
{III}result.emplace_back(current->Copy());
{II}}}
{II}// endregion

{II}// region Shift
{II}previous = current;
{II}previous_as_start = current_as_start;

{II}current = lookahead;
{II}current_as_start = lookahead_as_start;
{II}current_as_text = lookahead_as_text;

{II}if (current == nullptr) {{
{III}break;
{II}}}

{II}if (it == nodes.end()) {{
{III}lookahead = nullptr;
{III}lookahead_as_start = nullptr;
{III}lookahead_as_stop = nullptr;
{III}lookahead_as_text = nullptr;
{II}}} else {{
{III}lookahead = it->get();
{III}lookahead_as_start = dynamic_cast<StartNode*>(lookahead);
{III}lookahead_as_stop = dynamic_cast<StopNode*>(lookahead);
{III}lookahead_as_text = dynamic_cast<TextNode*>(lookahead);

{III}++it;
{II}}}
{II}// endregion
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            """\
}  // namespace canonicalizer"""
        ),
        Stripped(
            f"""\
std::string CanonicalizeXml(
{I}const std::string& xml
) {{
{I}const char namespace_separator = ':';
{I}XML_Parser parser = XML_ParserCreateNS(nullptr, namespace_separator);

{I}std::unique_ptr<canonicalizer::OurData> our_data(
{II}aas::common::make_unique<canonicalizer::OurData>(
{III}parser
{II})
{I});

{I}XML_SetUserData(parser, our_data.get());
{I}XML_SetElementHandler(
{II}parser,
{II}canonicalizer::OnStartElement,
{II}canonicalizer::OnStopElement
{I});
{I}XML_SetCharacterDataHandler(parser, canonicalizer::OnText);

{I}XML_Status status = XML_Parse(
{II}parser,
{II}xml.data(),
{II}static_cast<int>(xml.size()),
{II}true
{I});

{I}if (status == XML_STATUS_ERROR) {{
{II}XML_Error error_code = XML_GetErrorCode(parser);

{II}std::string message;

{II}if (error_code == XML_ERROR_ABORTED) {{
{III}if (!our_data->error.has_value()) {{
{IIII}throw std::logic_error(
{IIIII}"The XML parsing was aborted, "
{IIIII}"but the error in our_data was empty"
{IIII});
{III}}}

{III}message = std::move(*our_data->error);
{II}}} else {{
{III}const XML_LChar* error_str = XML_ErrorString(error_code);
{III}message = error_str;
{II}}}

{II}throw std::invalid_argument(
{III}aas::common::Concat(
{IIII}"Failed to parse XML for canonicalization: ",
{IIII}std::to_string(XML_GetCurrentLineNumber(parser)),
{IIII}":",
{IIII}std::to_string(XML_GetCurrentColumnNumber(parser)),
{IIII}": ",
{IIII}message
{III})
{II});
{I}}}

{I}std::deque<
{II}std::unique_ptr<canonicalizer::INode>
{I}> nodes(
{II}std::move(our_data->nodes)
{I});

{I}nodes = canonicalizer::MergeTextNodes(nodes);

{I}nodes = canonicalizer::RemoveNonSemanticWhiteSpace(nodes);

{I}size_t size = 0;
{I}for (const auto& node : nodes) {{
{II}size += node->text.size();
{I}}}

{I}std::string result;
{I}result.reserve(size);

{I}for (const auto& node : nodes) {{
{II}result.append(node->text);
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
std::shared_ptr<
{I}{library_namespace}::types::IClass
> MustDeserializeFile(
{I}const std::filesystem::path& path
) {{
{I}std::ifstream ifs(path, std::ios::binary);

{I}aas::common::expected<
{II}std::shared_ptr<aas::types::IClass>,
{II}aas::xmlization::DeserializationError
{I}> deserialized = aas::xmlization::From(
{II}ifs
{I});

{I}if (ifs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"The file stream is in the bad mode after "
{IIII}"reading and parsing the file as XML: ",
{IIII}path.string()
{III})
{II});
{I}}}

{I}if (!deserialized.has_value()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to de-serialize from ",
{IIII}path.string(),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(
{IIIII}deserialized.error().path.ToWstring()
{IIII}),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(
{IIIII}deserialized.error().cause
{IIII})
{III})
{II});
{I}}}

{I}return std::move(deserialized.value());
}}"""
        ),
        Stripped(
            """\
}  // namespace xmlization
}  // namespace common
}  // namespace test"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_header.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_header_consistent(
    module_doc=__doc__, generate_header_doc=generate_header.__doc__
)

assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the JSON de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test the JSON de/serialization of concrete classes."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_jsonization.hpp"

#include <{include_prefix_path}/jsonization.hpp>

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
template<class ClassT>
void AssertRoundTrip(
{I}const std::filesystem::path& path,
{I}std::function<
{II}aas::common::expected<
{III}std::shared_ptr<ClassT>,
{III}aas::jsonization::DeserializationError
{II}>(const nlohmann::json&, bool)
{I}> deserialization_function
) {{
{I}const nlohmann::json json = test::common::jsonization::MustReadJson(path);

{I}aas::common::expected<
{II}std::shared_ptr<ClassT>,
{II}aas::jsonization::DeserializationError
{I}> deserialized = deserialization_function(json, false);

{I}if (!deserialized.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Failed to de-serialize from ",
{IIII}path.string(),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(
{IIIII}deserialized.error().path.ToWstring()
{IIII}),
{IIII}": ",
{IIII}aas::common::WstringToUtf8(
{IIIII}deserialized.error().cause
{IIII})
{III})
{II})
{II}REQUIRE(deserialized.has_value());
{I}}}

{I}nlohmann::json another_json = aas::jsonization::Serialize(
{II}*(deserialized.value())
{I});

{I}std::optional<std::string> diff_message = test::common::jsonization::CompareJsons(
{II}json,
{II}another_json
{I});
{I}if (diff_message.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"The JSON round-trip from ",
{IIII}path.string(),
{IIII}" failed. There is a diff between the original JSON "
{IIII}"and the serialized one: ",
{IIII}*diff_message
{III})
{II})
{II}REQUIRE(!diff_message.has_value());
{I}}}
}}"""
        ),
        Stripped(
            f"""\
template<typename ClassT>
void AssertDeserializationFailure(
{I}const std::filesystem::path& path,
{I}std::function<
{II}aas::common::expected<
{III}std::shared_ptr<ClassT>,
{III}aas::jsonization::DeserializationError
{II}>(const nlohmann::json&, bool)
{I}> deserialization_function,
{I}const std::filesystem::path& error_path
) {{
{I}const nlohmann::json json = test::common::jsonization::MustReadJson(path);

{I}aas::common::expected<
{II}std::shared_ptr<ClassT>,
{II}aas::jsonization::DeserializationError
{I}> deserialized = deserialization_function(json, false);

{I}if (deserialized.has_value()) {{
{II}INFO(
{III}aas::common::Concat(
{IIII}"Expected the de-serialization to fail on ",
{IIII}path.string(),
{IIII}", but the de-serialization succeeded"
{III})
{II})
{II}REQUIRE(!deserialized.has_value());
{I}}}

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}aas::common::Concat(
{III}aas::common::WstringToUtf8(
{IIII}deserialized.error().path.ToWstring()
{III}),
{III}": ",
{III}aas::common::WstringToUtf8(
{IIII}deserialized.error().cause
{III})
{II}),
{II}error_path
{I});
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineJsonDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "Json";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineErrorDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "JsonizationError";
{I}}}

{I}return *result;
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        interface_name = cpp_naming.interface_name(concrete_cls.name)

        model_type = naming.json_model_type(concrete_cls.name)

        deserialization_function = cpp_naming.function_name(
            Identifier(f"{concrete_cls.name}_from")
        )

        cls_name = cpp_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test the round-trip of an expected {cls_name}") {{
{I}const std::deque<std::filesystem::path> paths(
{II}test::common::FindFilesBySuffixRecursively(
{III}DetermineJsonDir()
{IIII}/ "Expected"
{IIII}/ {cpp_common.string_literal(model_type)},
{III}".json"
{II})
{I});

{I}for (const std::filesystem::path& path : paths) {{
{II}AssertRoundTrip<
{III}aas::types::{interface_name}
{II}>(path, aas::jsonization::{deserialization_function});
{I}}}
}}"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test the de-serialization failure on an unexpected {cls_name}") {{
{I}for (
{II}const std::filesystem::path& causeDir
{II}: test::common::ListSubdirectories(
{III}DetermineJsonDir()
{IIII}/ "Unexpected"
{IIII}/ "Unserializable"
{II})
{I}) {{
{II}for (
{III}const std::filesystem::path& path
{III}: test::common::FindFilesBySuffixRecursively(
{IIII}causeDir / {cpp_common.string_literal(model_type)},
{IIII}".json"
{III})
{II}) {{
{III}const std::filesystem::path parent(
{IIII}(
{IIIII}DetermineErrorDir()
{IIIII}/ std::filesystem::relative(path, DetermineJsonDir())
{IIII}).parent_path()
{III});

{III}const std::filesystem::path error_path(
{IIII}parent
{IIII}/ (path.filename().string() + ".error")
{III});

{III}AssertDeserializationFailure<
{IIII}aas::types::{interface_name}
{III}>(
{IIII}path,
{IIII}aas::jsonization::{deserialization_function},
{IIII}error_path
{III});
{II}}}
{I}}}
}}"""
            )
        )

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the XML de/serialization of interfaces."""

from typing import List

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the XML de/serialization of interfaces.
    """
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            interface_name_java = java_naming.interface_name(cls.interface.name)

            implementer_cls_name_java = java_naming.class_name(implementer_cls.name)

            blocks.append(
                Stripped(
                    f"""\
@Test
public void testRoundTrip{interface_name_java}From{implementer_cls_name_java}()
{I}throws IOException, XMLStreamException {{
{I}// We load from JSON here just to jump-start the round trip.
{I}// The round-trip goes then over XML.
{I}final {interface_name_java} instance =
{II}CommonJsonization.loadMaximal{implementer_cls_name_java}();

{I}// The round-trip starts here.
{I}final StringWriter stringOut = new StringWriter();
{I}final XMLOutputFactory outputFactory = XMLOutputFactory.newFactory();
{I}final XMLStreamWriter xmlWriter = outputFactory.createXMLStreamWriter(stringOut);

{I}Xmlization.Serialize.to(instance, xmlWriter);
{I}String outputText = stringOut.toString();

{I}// De-serialize from XML
{I}final XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
{I}final XMLEventReader xmlReader =
{II}xmlInputFactory.createXMLEventReader(new StringReader(outputText));
{I}final {interface_name_java} anotherInstance =
{II}Xmlization.Deserialize.deserialize{interface_name_java}(xmlReader);

{I}// Serialize back to XML
{I}final StringWriter anotherStringOut = new StringWriter();
{I}final XMLStreamWriter anotherXmlWriter = outputFactory.createXMLStreamWriter(anotherStringOut);
{I}Xmlization.Serialize.to(anotherInstance, anotherXmlWriter);

{I}// Compare
{I}assertEquals(outputText, anotherStringOut.toString());
}} // void testRoundTrip{interface_name_java}From{implementer_cls_name_java}"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestXmlizationOfInterfaces.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.types.impl.*;
import {package}.types.model.*;
import {package}.xmlization.Xmlization;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import javax.xml.stream.*;
import org.junit.jupiter.api.Test;

public class TestXmlizationOfInterfaces {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestXmlizationOfInterfaces

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the virtual machine for matching regular expressions."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate implementation to test the virtual machine for matching regular expressions."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            """\
/**
 * Test the virtual machine for matching regular expressions.
 */"""
        ),
        Stripped(
            f'''\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/revm.hpp"'''
        ),
        Stripped(
            """\
#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>"""
        ),
        Stripped(
            f"""\
namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test character in an empty range") {{
{I}std::vector<aas::revm::Range> ranges;

{I}REQUIRE(
{II}!aas::revm::CharacterInRanges(ranges, L'M')
{I});
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test character in a single range") {{
{I}std::vector<aas::revm::Range> ranges = {{
{II}aas::revm::Range(L'A', L'Z')
{I}}};

{I}REQUIRE(CharacterInRanges(ranges, L'M'));
{I}REQUIRE(!CharacterInRanges(ranges, L'a'));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test character in multiple ranges") {{
{I}std::vector<aas::revm::Range> ranges = {{
{II}aas::revm::Range(L'A', L'C'),
{II}aas::revm::Range(L'D', L'F'),
{II}aas::revm::Range(L'G', L'I'),
{II}aas::revm::Range(L'J', L'L'),
{II}aas::revm::Range(L'M', L'N')
{I}}};

{I}for (wchar_t character = L'A'; character < L'Z'; ++character) {{
{II}const std::string message(
{III}aas::common::WstringToUtf8(
{IIII}L"Testing for character: " + std::wstring(1, character)
{III})
{II});

{II}if (character <= L'N') {{
{III}INFO(message)
{III}REQUIRE(
{IIII}CharacterInRanges(ranges, character)
{III});
{II}}} else {{
{III}INFO(message)
{III}REQUIRE(
{IIII}!CharacterInRanges(ranges, character)
{III});
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^a$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^a+b+$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSplit>(0, 2)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'b')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSplit>(2, 4)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(!aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(aas::revm::Match(program, L"ab"));
{I}REQUIRE(aas::revm::Match(program, L"aabb"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^a|b$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSplit>(1, 3)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionJump>(4)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'b')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^a?$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSplit>(1, 2)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^a*$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSplit>(1, 3)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionJump>(0)
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^[a-b]$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionSet>(
{III}std::vector<aas::revm::Range>{{
{IIII}aas::revm::Range(L'a', L'b')
{III}}}
{II})
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^[^a-b]$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionNotSet>(
{III}std::vector<aas::revm::Range>{{
{IIII}aas::revm::Range(L'a', L'b')
{III}}}
{II})
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionEnd>()
{I});
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{I});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(!aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
{I}REQUIRE(aas::revm::Match(program, L"c"));
{I}REQUIRE(!aas::revm::Match(program, L"cc"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern ^.$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(std::make_unique<aas::revm::InstructionAny>());
{I}program.emplace_back(std::make_unique<aas::revm::InstructionEnd>());
{I}program.emplace_back(std::make_unique<aas::revm::InstructionMatch>());

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(!aas::revm::Match(program, L"aa"));
{I}REQUIRE(aas::revm::Match(program, L"b"));
{I}REQUIRE(!aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        Stripped(
            f"""\
TEST_CASE("Test pattern with arbitrary suffix ^a.*$") {{
{I}std::vector<std::unique_ptr<aas::revm::Instruction> > program;

{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionChar>(L'a')
{III});

{I}// NOTE (mristin):
{I}// Usually, the last two instructions would be `end` followed by a `match`.
{I}// However, we can optimize by having a single `match` without an `end` so that
{I}// we return early.
{I}program.emplace_back(
{II}std::make_unique<aas::revm::InstructionMatch>()
{III});

{I}REQUIRE(!aas::revm::Match(program, L""));
{I}REQUIRE(aas::revm::Match(program, L"a"));
{I}REQUIRE(aas::revm::Match(program, L"aa"));
{I}REQUIRE(!aas::revm::Match(program, L"b"));
{I}REQUIRE(aas::revm::Match(program, L"ab"));
{I}REQUIRE(!aas::revm::Match(program, L"ba"));
}}"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code for common JSON de/serialization shared across the tests."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.cpp import common as cpp_common
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(library_namespace: Stripped) -> str:
    """Generate header for common JSON de/serialization shared across the tests."""
    include_guard_var = cpp_common.include_guard_var(
        Stripped("test::common::jsonization")
    )

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
/**
 * Provide methods which are used throughout the jsonization tests.
 */
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        Stripped(
            """\
#include <nlohmann/json.hpp>"""
        ),
        Stripped(
            """\
#include <optional>
#include <vector>
#include <string>"""
        ),
        Stripped(
            """\
namespace test {
namespace common {
namespace jsonization {"""
        ),
        Stripped(
            """\
/**
 * Read the content of the `path` and parse it as JSON.
 */
nlohmann::json MustReadJson(const std::filesystem::path& path);"""
        ),
        Stripped(
            f"""\
/**
 * Compare two JSON values.
 *
 * Return the JSON patch, if there is any difference.
 */
std::optional<std::string> CompareJsons(
{I}const nlohmann::json& that,
{I}const nlohmann::json& other
);"""
        ),
        Stripped(
            """\
}  // namespace jsonization
}  // namespace common
}  // namespace test"""
        ),
        Stripped(
            f"""\
#endif  // {include_guard_var}"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(library_namespace: Stripped) -> str:
    """Generate implementation for common JSON de/serialization shared across the tests."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            '''\
#include "./common_jsonization.hpp"'''
        ),
        Stripped(
            f"""\
#include <{include_prefix_path}/common.hpp>"""
        ),
        Stripped(
            """\
#include <fstream>"""
        ),
        Stripped(
            f"""\
namespace aas = {library_namespace};"""
        ),
        Stripped(
            """\
namespace test {
namespace common {
namespace jsonization {"""
        ),
        Stripped(
            f"""\
nlohmann::json MustReadJson(const std::filesystem::path& path) {{
{I}std::ifstream ifs(path);

{I}nlohmann::json result;
{I}try {{
{II}result = nlohmann::json::parse(ifs);
{I}}} catch (nlohmann::json::parse_error& exception) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read JSON from ",
{IIII}path.string(),
{IIII}": ",
{IIII}exception.what()
{III})
{II});
{I}}}

{I}if (ifs.bad()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read JSON from ",
{IIII}path.string(),
{IIII}"; the bad bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (ifs.fail()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read JSON from ",
{IIII}path.string(),
{IIII}"; the fail bit of the file stream is set"
{III})
{II});
{I}}}

{I}if (!ifs.eof()) {{
{II}throw std::runtime_error(
{III}aas::common::Concat(
{IIII}"Failed to read JSON from ",
{IIII}path.string(),
{IIII}"; the EOF bit is not set meaning that we did not parse all the content"
{III})
{II});
{I}}}

{I}return result;
}}"""
        ),
        Stripped(
            f"""\
std::optional<std::string> CompareJsons(
{I}const nlohmann::json& that,
{I}const nlohmann::json& other
) {{
{I}const nlohmann::json patch = nlohmann::json::diff(that, other);

{I}if (!patch.is_array()) {{
{II}throw std::logic_error(
{III}aas::common::Concat(
{IIII}"Expected the patch to be an array, but got ",
{IIII}patch.type_name()
{III})
{II});
{I}}}

{I}if (patch.size() != 0) {{
{II}return nlohmann::to_string(patch);
{I}}}

{I}return std::nullopt;
}}"""
        ),
        Stripped(
            """\
}  // namespace jsonization
}  // namespace common
}  // namespace test"""
        ),
        cpp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_header.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_header_consistent(
    module_doc=__doc__, generate_header_doc=generate_header.__doc__
)

assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the ``OverXOrEmpty`` methods."""

from typing import List

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line, Identifier
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the ``OverXOrEmpty`` methods.
    """
    blocks = []  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)

        for prop in concrete_cls.properties:
            method_name_java = java_naming.method_name(
                Identifier(f"over_{prop.name}_or_empty")
            )
            method_name_capitalized_java = (
                method_name_java[0].upper() + method_name_java[1:]
            )
            getter_name_java = java_naming.getter_name(prop.name)

            if isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ) and isinstance(
                prop.type_annotation.value, intermediate.ListTypeAnnotation
            ):
                blocks.append(
                    Stripped(
                        f"""\
@Test
public void test{cls_name_java}{method_name_capitalized_java}() throws IOException {{
{I}for ({cls_name_java} instance : new {cls_name_java}[]
{I}{{
{II}CommonJsonization.loadMinimal{cls_name_java}(),
{II}CommonJsonization.loadMaximal{cls_name_java}()
{I}}}) {{
{II}int length = instance.{getter_name_java}().map(elem -> elem.size()).orElse(0);
{II}AtomicInteger count = new AtomicInteger();
{II}instance.{method_name_java}().forEach(i -> count.getAndIncrement());
{II}assertEquals(length , count.get());
{I}}}
}} // public void test{cls_name_java}{method_name_java}"""
                    )
                )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestOverXOrEmpty.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.types.enums.*;
import {package}.types.impl.*;
import java.io.IOException;
import java.util.concurrent.atomic.AtomicInteger;
import org.junit.jupiter.api.Test;

public class TestOverXOrEmpty {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestOverXOrEmpty

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the test code for the jsonization of classes with descendants."""

import io

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import INDENT as I, INDENT2 as II, INDENT3 as III


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate the test code for the jsonization of classes with descendants."""
    blocks = [
        Stripped("package jsonization_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"testing"
{I}aasjsonization "{repo_url}/jsonization"
{I}aastesting "{repo_url}/aastesting"
)"""
        ),
    ]

    for cls in symbol_table.classes:
        # NOTE (mristin):
        # We can only de-serialize instances of concrete descendants which carry
        # a model type in their serializations.
        concrete_descendants_with_model_type = [
            concrete_descendant
            for concrete_descendant in cls.concrete_descendants
            if concrete_descendant.serialization.with_model_type
        ]

        if len(concrete_descendants_with_model_type) == 0:
            continue

        descendant_cls = concrete_descendants_with_model_type[0]

        must_load_minimal_name = golang_naming.function_name(
            Identifier(f"must_load_minimal_{descendant_cls.name}")
        )

        deserialization_function = golang_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        test_name = golang_naming.function_name(
            (Identifier(f"test_{cls.name}_round_trip_OK_over_descendant"))
        )

        model_type = naming.json_model_type(descendant_cls.name)

        blocks.append(
            Stripped(
                f"""\
func {test_name}(t *testing.T) {{
{I}instance := aastesting.{must_load_minimal_name}()

{I}jsonable, err := aasjsonization.ToJsonable(instance)
{I}if err != nil {{
{II}t.Fatalf(
{III}"Failed to serialize the minimal {model_type}: %v",
{III}err,
{II})
{II}return
{I}}}

{I}source := "<minimal {model_type}>"

{I}deserialized, deseriaErr := aasjsonization.{deserialization_function}(
{II}jsonable,
{I})
{I}ok := assertNoDeserializationError(t, deseriaErr, source)
{I}if !ok {{
{II}return
{I}}}

{I}anotherJsonable, seriaErr := aasjsonization.ToJsonable(deserialized)
{I}ok = assertNoSerializationError(t, seriaErr, source)
{I}if !ok {{
{II}return
{I}}}

{I}ok = assertSerializationEqualsDeserialization(
{II}t,
{II}jsonable,
{II}anotherJsonable,
{II}source,
{I})
{I}if !ok {{
{II}return
{I}}}
}}"""
            )
        )

        # NOTE (mristin):
        # We test here only abstract classes as the concrete classes are going
        # to be already tested in another test with concrete classes.
        if isinstance(cls, intermediate.AbstractClass):
            test_name = golang_naming.function_name(
                (Identifier(f"test_{cls.name}_deserialization_fail"))
            )

            blocks.append(
                Stripped(
                    f"""\
func {test_name}(t *testing.T) {{
{I}jsonable := any("this is not an object")

{I}_, err := aasjsonization.{deserialization_function}(
{II}jsonable,
{I})

{I}if err == nil {{
{II}t.Fatal("Expected an error, but got none.")
{II}return
{I}}}

{I}deseriaErr, ok := err.(*aasjsonization.DeserializationError)
{I}if !ok {{
{II}t.Fatalf("Expected a de-serialization error, but got: %v", err)
{II}return
{I}}}

{I}pathString := deseriaErr.PathString()
{I}if len(pathString) != 0 {{
{II}t.Fatalf(
{III}"Expected an empty path in error, but got: %s",
{III}pathString,
{II})
{II}return
{I}}}

{I}expectedMessage :=
{II}"Expected a JSON object, but got string"

{I}if deseriaErr.Message != expectedMessage {{
{II}t.Fatalf(
{III}"Expected the deserialization error:\\n%s\\n, but got:\\n%s",
{III}expectedMessage, deseriaErr.Message,
{II})
{II}return
{I}}}
}}"""
                )
            )

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the wstringification of enums."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


def _generate_test_round_trip_for_model_type(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """
    Generate the test case for valid de/wstringification of model type enum.

    This enum does not explicitly exist in the meta-model, so we have to generate
    the test for it separately.
    """
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    statements = []  # type: List[Stripped]

    must_from_wstring = cpp_naming.function_name(
        Identifier("must_model_type_from_wstring")
    )

    for cls in symbol_table.concrete_classes:
        literal_name = cpp_naming.enum_literal_name(cls.name)
        literal_value = naming.json_model_type(cls.name)

        statements.append(
            Stripped(
                f"""\
REQUIRE(
{I}aas::types::{enum_name}::{literal_name}
{I}== aas::wstringification::{must_from_wstring}(
{II}{cpp_common.wstring_literal(literal_value)}
{I})
);"""
            )
        )

        statements.append(
            Stripped(
                f"""\
REQUIRE(
{I}aas::wstringification::to_wstring(
{II}aas::types::{enum_name}::{literal_name}
{I})
{I}== {cpp_common.wstring_literal(literal_value)}
);"""
            )
        )

    body = "\n\n".join(statements)

    return Stripped(
        f"""\
TEST_CASE("Test {enum_name} round-trip") {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_test_failure_for_model_type() -> Stripped:
    """
    Generate the test case for invalid de-wstringification of model type enum.

    This enum does not explicitly exist in the meta-model, so we have to generate
    the test for it separately.
    """
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    from_wstring = cpp_naming.function_name(Identifier("model_type_from_wstring"))

    must_from_wstring = cpp_naming.function_name(
        Identifier("must_model_type_from_wstring")
    )

    return Stripped(
        f"""\
TEST_CASE("Test failure on {enum_name}") {{
{I}CHECK(
{II}!aas::wstringification::{from_wstring}(
{III}L"Totally utterly invalid"
{II}).has_value()
{I});

{I}REQUIRE_THROWS_WITH(
{II}aas::wstringification::{must_from_wstring}(
{III}L"Totally utterly invalid"
{II}),
{II}"Unexpected {enum_name} literal: Totally utterly invalid"
{I});
}}"""
    )


def _generate_test_round_trip_for(enum: intermediate.Enumeration) -> Stripped:
    """Generate the test of the valid de/serialization for an enum."""
    enum_name = cpp_naming.enum_name(enum.name)

    statements = []  # type: List[Stripped]

    must_from_wstring = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_wstring")
    )

    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)

        statements.append(
            Stripped(
                f"""\
REQUIRE(
{I}aas::types::{enum_name}::{literal_name}
{I}== aas::wstringification::{must_from_wstring}(
{II}{cpp_common.wstring_literal(literal.value)}
{I})
);"""
            )
        )

        statements.append(
            Stripped(
                f"""\
REQUIRE(
{I}aas::wstringification::to_wstring(
{II}aas::types::{enum_name}::{literal_name}
{I})
{I}== {cpp_common.wstring_literal(literal.value)}
);"""
            )
        )

    body = "\n\n".join(statements)

    return Stripped(
        f"""\
TEST_CASE("Test {enum_name} round-trip") {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_test_failure_for(enum: intermediate.Enumeration) -> Stripped:
    """Generate the de-wstringification of an invalid literal."""
    enum_name = cpp_naming.enum_name(enum.name)

    from_wstring = cpp_naming.function_name(Identifier(f"{enum.name}_from_wstring"))

    must_from_wstring = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_wstring")
    )

    return Stripped(
        f"""\
TEST_CASE("Test failure on {enum_name}") {{
{I}CHECK(
{II}!aas::wstringification::{from_wstring}(
{III}L"Totally utterly invalid"
{II}).has_value()
{I});

{I}REQUIRE_THROWS_WITH(
{II}aas::wstringification::{must_from_wstring}(
{III}L"Totally utterly invalid"
{II}),
{II}"Unexpected {enum_name} literal: Totally utterly invalid"
{I});
}}"""
    )


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test the wstringification of enums."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/wstringification.hpp"

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace aas = {library_namespace};"""
        ),
        _generate_test_round_trip_for_model_type(symbol_table=symbol_table),
        _generate_test_failure_for_model_type(),
    ]  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        if len(enum.literals) == 0:
            continue

        blocks.append(_generate_test_round_trip_for(enum=enum))

        blocks.append(_generate_test_failure_for(enum=enum))

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code for common XML de/serialization shared across the tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(repo_url: Stripped) -> str:
    """Generate code for common XML de/serialization shared across the tests."""
    blocks = [
        Stripped(
            """\
package xmlization_test"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"os"
{I}"path/filepath"
{I}"regexp"
{I}"strings"
{I}"testing"
{I}aastesting "{repo_url}/aastesting"
{I}aasxmlization "{repo_url}/xmlization"
)"""
        ),
        Stripped(
            f"""\
// Assert that there is no de-serialization error when de-serializing
// from `source`.
func assertNoDeserializationError(
{I}t *testing.T,
{I}err error,
{I}source string,
) (ok bool) {{
{I}ok = true
{I}if err != nil {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected no de-serialization error from %s, but got: %s",
{III}source, err.Error(),
{II})
{II}return
{I}}}
{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that there is no serialization error when serializing the instance
// originally coming from `source`.
func assertNoSerializationError(
{I}t *testing.T,
{I}err error,
{I}source string,
) (ok bool) {{
{I}ok = true
{I}if err != nil {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected no serialization error when serializing "+
{IIII}"the instance obtained from %s, but got: %s",
{III}source, err.Error(),
{II})
{II}return
{I}}}
{I}return
}}"""
        ),
        Stripped(
            """\
// NOTE (mristin):
// Currently, Go does not support self-closing tags,
// see: https://github.com/golang/go/issues/21399.
// We apply the following hack to make the tags self-closing even if they are not.
// This is unsafe in general, but works OK for the limited set of test data that we are
// here dealing with.
//
// The code has been taken from: https://github.com/golang/go/issues/21399#issuecomment-1342730174"""
        ),
        Stripped(
            """\
var emptyTagRe = regexp.MustCompile(`<(\\w+)></\\w+>`)"""
        ),
        Stripped(
            f"""\
func forceSelfClosingTags(text string) string {{
{I}b := []byte(text)
{I}emptyTagIdxs := emptyTagRe.FindAllSubmatchIndex(b, -1)

{I}if len(emptyTagIdxs) == 0 {{
{II}return text
{I}}}

{I}var nb []byte

{I}for _, idx := range emptyTagIdxs {{
{II}// Get everything in b up till the first of the submatch indexes (this is
{II}// the start of an "empty" <thing></thing> tag), then get the name of the tag
{II}// and put it in a self-closing tag.
{II}nb = append(b[0:idx[0]], fmt.Sprintf("<%s/>", b[idx[2]:idx[3]])...)

{II}// Finally, append everything *after* the submatch indexes
{II}nb = append(nb, b[len(b)-(len(b)-idx[1]):]...)
{I}}}

{I}return string(nb)
}}"""
        ),
        Stripped(
            """\
var whitespaceBetweenTagsRe = regexp.MustCompile(`>\\s+<`)"""
        ),
        Stripped(
            f"""\
// Remove all whitespace (including newlines, tabs, and spaces) between XML tags
// without parsing the XML.
func removeWhitespaceBetweenTags(text string) string {{
{I}return whitespaceBetweenTagsRe.ReplaceAllString(text, "><")
}}"""
        ),
        Stripped(
            """\
var whitespaceBeforeAngleBracketsRe = regexp.MustCompile(`\\s+>`)
var whitespaceBeforeSelfClosingRe = regexp.MustCompile(`\\s+/>`)
var whitespaceAfterAngleBracketsRe = regexp.MustCompile(`<\\s+`)"""
        ),
        Stripped(
            f"""\
// Remove any whitespace before `>` and `/>` or after `<`
func removeWhitespaceWithinTags(text string) string {{
{I}result := text
{I}result = whitespaceBeforeAngleBracketsRe.ReplaceAllString(result, ">")
{I}result = whitespaceBeforeSelfClosingRe.ReplaceAllString(result, "/>")
{I}result = whitespaceAfterAngleBracketsRe.ReplaceAllString(result, "<")
{I}return result
}}"""
        ),
        Stripped(
            f"""\
// Assert that the serialization `other`, as XML document, equals the original
//
//	XML document `that` read from the `source`.
func assertSerializationEqualsDeserialization(
{I}t *testing.T,
{I}that string,
{I}other string,
{I}source string,
) (ok bool) {{
{I}// Remove carriers to avoid problems between Windows, Posix and MacOS
{I}canonicalThat := strings.ReplaceAll(that, "\\r", "")
{I}canonicalOther := strings.ReplaceAll(other, "\\r", "")

{I}canonicalThat = strings.TrimSpace(canonicalThat)
{I}canonicalOther = strings.TrimSpace(canonicalOther)

{I}canonicalThat = forceSelfClosingTags(canonicalThat)
{I}canonicalOther = forceSelfClosingTags(canonicalOther)

{I}// NOTE (mristin):
{I}// The following hack is SUPER ugly and unsafe! However, it works. Given Go's
{I}// limited support for XML, we gave up on a safer approach :(. We tested
{I}// the following approaches before applying this hack:
{I}//  * A round-trip over `encoding/xml`. Failed due to
{I}//    https://github.com/golang/go/issues/13400.
{I}//  * Using `aqwari.net/xml/xmltree`. Failed as the special characters in the
{I}//    element content still has not been de-escaped or consistently escaped in
{I}//    a round trip.

{I}canonicalThat = strings.ReplaceAll(canonicalThat, "'", "&#39;")
{I}canonicalOther = strings.ReplaceAll(canonicalOther, "'", "&#39;")

{I}canonicalThat = removeWhitespaceBetweenTags(canonicalThat)
{I}canonicalOther = removeWhitespaceBetweenTags(canonicalOther)

{I}canonicalThat = removeWhitespaceWithinTags(canonicalThat)
{I}canonicalOther = removeWhitespaceWithinTags(canonicalOther)

{I}thatLines := strings.Split(canonicalThat, "\\n")
{I}otherLines := strings.Split(canonicalOther, "\\n")

{I}if canonicalThat != canonicalOther {{
{II}b := new(strings.Builder)
{II}minLines := len(thatLines)
{II}if minLines > len(otherLines) {{
{III}minLines = len(otherLines)
{II}}}
{II}for i := 0; i < minLines; i++ {{
{III}if thatLines[i] == otherLines[i] {{
{IIII}b.WriteString(fmt.Sprintf("           %s\\n", thatLines[i]))
{III}}} else {{
{IIII}b.WriteString(fmt.Sprintf("ORIGINAL   %s\\n", thatLines[i]))
{IIII}b.WriteString(fmt.Sprintf("SERIALIZED %s\\n", otherLines[i]))
{IIII}break
{III}}}
{II}}}

{II}ok = false
{II}t.Fatalf(
{III}"The canonicalized XML serialization of the de-serialized instance "+
{IIII}"from %s does not equal the canonicalized original XML document:\\n"+
{IIII}"%s",
{III}source, b.String(),
{II})
{II}return
{I}}}

{I}return
}}"""
        ),
        Stripped(
            f"""\
// Assert that there is a de-serialization error.
//
// If [aastesting.RecordMode] is set, the de-serialization error is re-recorded
// to `expectedPth`. Otherwise, the error is compared against the golden file
// `expectedPth`.
func assertIsDeserializationErrorAndEqualsExpectedOrRecord(
{I}t *testing.T,
{I}err error,
{I}source string,
{I}expectedPth string,
) (ok bool) {{
{I}ok = true

{I}if err == nil {{
{II}ok = false
{II}t.Fatalf("De-serialization error expected from %s, but got none", source)
{II}return
{I}}}

{I}deseriaErr, is := err.(*aasxmlization.DeserializationError)
{I}if !is {{
{II}ok = false
{II}t.Fatalf(
{III}"Expected a de-serialization error, "+
{IIII}"but got an error of type %T from %s: %v",
{III}err, source, err,
{II})
{II}return
{I}}}

{I}// Add a new line for POSIX systems.
{I}got := deseriaErr.Error() + "\\n"

{I}if aastesting.RecordMode {{
{II}parent := filepath.Dir(expectedPth)
{II}err := os.MkdirAll(parent, os.ModePerm)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to create the directory %s: %s", parent, err.Error(),
{IIII}),
{III})
{II}}}

{II}err = os.WriteFile(expectedPth, []byte(got), 0644)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to write to the file %s: %s", expectedPth, err.Error(),
{IIII}),
{III})
{II}}}
{I}}} else {{
{II}_, err := os.Stat(expectedPth)
{II}if err != nil {{
{III}ok = false
{III}t.Fatalf(
{IIII}"Failed to stat the file %s: %s; if the file does not exist, "+
{IIIII}"you probably want to record the test data by "+
{IIIII}"setting the environment variable %s",
{IIII}expectedPth, err.Error(), aastesting.RecordModeEnvironmentVariableName,
{III})
{III}return
{II}}}

{II}bb, err := os.ReadFile(expectedPth)
{II}if err != nil {{
{III}panic(
{IIII}fmt.Sprintf(
{IIIII}"Failed to read from file %s: %s", expectedPth, err.Error(),
{IIII}),
{III})
{II}}}

{II}expected := string(bb)

{II}// NOTE (mristin):
{II}// Git automatically strips and adds `\\r`, so we have to remove it here
{II}// to obtain a canonical text.
{II}expected = strings.Replace(expected, "\\r", "", -1)

{II}if expected != got {{
{III}ok = false
{III}t.Fatalf(
{IIII}"What we got differs from the expected in %s. "+
{IIIII}"We got:\\n%s\\nWe expected:\\n%s",
{IIII}expectedPth, got, expected,
{III})
{III}return
{II}}}
{I}}}

{I}return
}}"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test ``Descent`` and ``DescentOnce`` iterations."""


import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate implementation to test ``Descent`` and ``DescentOnce`` iterations."""
    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "./common.hpp"
#include "./common_examples.hpp"
#include "./common_xmlization.hpp"

#include <{include_prefix_path}/iteration.hpp>
#include <{include_prefix_path}/stringification.hpp>
#include <{include_prefix_path}/types.hpp>

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

#include <deque>

namespace aas = {library_namespace};"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineXmlDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "Xml";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineDescentDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "Descent";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
const std::filesystem::path& DetermineDescentOnceDir() {{
{I}static aas::common::optional<std::filesystem::path> result;
{I}if (!result.has_value()) {{
{II}result = test::common::DetermineTestDataDir() / "DescentOnce";
{I}}}

{I}return *result;
}}"""
        ),
        Stripped(
            f"""\
void AssertOrRerecordDescent(
{I}const std::filesystem::path& xml_path,
{I}const std::filesystem::path& trace_path
) {{
{I}std::shared_ptr<
{II}aas::types::IClass
{I}> instance(
{II}test::common::xmlization::MustDeserializeFile(xml_path)
{I});

{I}std::deque<std::wstring> parts;

{I}for (
{II}const std::shared_ptr<aas::types::IClass>& something
{II}: aas::iteration::Descent(instance)
{I}) {{
{II}parts.emplace_back(test::common::TraceMark(*something));
{II}parts.push_back(L"\\n");
{I}}}

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}aas::common::WstringToUtf8(
{III}test::common::JoinWstrings(parts, L"")
{II}),
{II}trace_path
{I});
}}"""
        ),
        Stripped(
            f"""\
void AssertOrRerecordDescentOnce(
{I}const std::shared_ptr<aas::types::IClass>& instance,
{I}const std::filesystem::path& trace_path
) {{
{I}std::deque<std::wstring> parts;

{I}for (
{II}const std::shared_ptr<aas::types::IClass>& something
{II}: aas::iteration::DescentOnce(instance)
{I}) {{
{II}parts.emplace_back(test::common::TraceMark(*something));
{II}parts.push_back(L"\\n");
{I}}}

{I}test::common::AssertContentEqualsExpectedOrRecord(
{II}aas::common::WstringToUtf8(
{III}test::common::JoinWstrings(parts, L"")
{II}),
{II}trace_path
{I});
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        xml_class_name = naming.xml_class_name(concrete_cls.name)

        cls_name = cpp_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test Descent over an {cls_name}") {{
{I}const std::deque<std::filesystem::path> paths(
{II}test::common::FindFilesBySuffixRecursively(
{III}DetermineXmlDir()
{IIII}/ "Expected"
{IIII}/ {cpp_common.string_literal(xml_class_name)},
{III}".xml"
{II})
{I});

{I}for (const std::filesystem::path& path : paths) {{
{II}const std::filesystem::path parent(
{III}(
{IIII}DetermineDescentDir()
{IIIII}/ std::filesystem::relative(path, DetermineXmlDir())
{III}).parent_path()
{II});

{II}const std::filesystem::path trace_path(
{III}parent
{IIII}/ (path.filename().string() + ".trace")
{II});

{II}AssertOrRerecordDescent(path, trace_path);
{I}}}
}}"""
            )
        )

        interface_name = cpp_naming.interface_name(concrete_cls.name)
        load_max = cpp_naming.function_name(Identifier(f"load_max_{concrete_cls.name}"))

        blocks.append(
            Stripped(
                f"""\
TEST_CASE("Test DescentOnce over an {cls_name}") {{
{I}const std::shared_ptr<
{II}aas::types::{interface_name}
{I}> instance(
{II}test::common::examples::{load_max}()
{I});

{I}const std::filesystem::path trace_path(
{II}DetermineDescentOnceDir()
{III}/ "Max{cls_name}.trace"
{I});

{I}AssertOrRerecordDescentOnce(instance, trace_path);
}}"""
            )
        )

    blocks.append(cpp_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate_implementation.__doc__ is not None
cpp_common.assert_module_docstring_and_generate_implementation_consistent(
    module_doc=__doc__, generate_implementation_doc=generate_implementation.__doc__
)



"""Generate code to test the JSON de/serialization of interfaces."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the JSON de/serialization of interfaces."""
    blocks = [
        Stripped(
            """\
/**
 * Test JSON de/serialization of interfaces.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasJsonization from "../src/jsonization";
import * as TestCommon from "./common";
import * as TestCommonJsonization from "./commonJsonization";"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        interface_name_typescript = typescript_naming.interface_name(cls.name)

        deserialization_function = typescript_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            cls_name_typescript = typescript_naming.class_name(implementer_cls.name)

            load_minimal_name = typescript_naming.function_name(
                Identifier(f"load_minimal_{implementer_cls.name}")
            )

            blocks.append(
                Stripped(
                    f"""\
test(
{I}"{interface_name_typescript} round-trip " +
{I}"starting from {cls_name_typescript} OK",
{I}() => {{
{I}const instance = TestCommonJsonization.{load_minimal_name}();

{I}const jsonable = AasJsonization.toJsonable(instance);

{I}const anotherInstanceOrError = AasJsonization.{deserialization_function}(
{II}jsonable
{I});
{I}expect(anotherInstanceOrError.error).toBeNull();
{I}const anotherInstance = anotherInstanceOrError.mustValue();

{I}const anotherJsonable = AasJsonization.toJsonable(anotherInstance);

{I}const inequalityError = TestCommon.checkJsonablesEqual(
{II}jsonable,
{II}anotherJsonable
{I});
{I}if (inequalityError !== null) {{
{II}throw new Error(
{III}"The minimal example of {interface_name_typescript} " +
{III}"as an instance of {cls_name_typescript} serialized " +
{III}"to JSON, then de-serialized and serialized again does not match " +
{III}`the first JSON: ${{inequalityError.path}}: ${{inequalityError.message}}`
{II});
{I}}}
}});"""
                )
            )

        blocks.append(
            Stripped(
                f"""\
test("{interface_name_typescript} deserialization fail", () => {{
{I}const jsonable = "This is not a {interface_name_typescript}.";

{I}const instanceOrError = AasJsonization.{deserialization_function}(
{II}jsonable
{I});
{I}expect(instanceOrError.error.message).toStrictEqual(
{II}"Expected a JSON object, but got: string"
{I});
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test ``DescendOnce*`` functions."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
)


def _generate_for_cls(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the test function."""
    test_function_name = golang_naming.function_name(
        Identifier(f"test_descend_once_on_an_instance_of_{cls.name}")
    )

    must_load_maximal_name = golang_naming.function_name(
        Identifier(f"must_load_maximal_{cls.name}")
    )

    model_type_literal = golang_common.string_literal(naming.json_model_type(cls.name))

    return Stripped(
        f"""\
func {test_function_name}(
{I}t *testing.T,
) {{
{I}instance := aastesting.{must_load_maximal_name}()

{I}expectedPth := filepath.Join(
{II}aastesting.TestDataDir,
{II}"DescendOnce",
{II}{model_type_literal},
{II}"maximal.json.trace",
{I})

{I}onlyOnce := true

{I}message := compareOrRerecordTrace(
{II}instance,
{II}expectedPth,
{II}onlyOnce,
{I})
{I}if message != nil {{
{II}t.Fatal(*message)
{I}}}
}}"""
    )


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test ``DescendOnce*`` functions."""
    blocks = [
        Stripped("package types_descend_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"path/filepath"
{I}"testing"
{I}aastesting "{repo_url}/aastesting"
)"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_for_cls(cls=concrete_cls))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code for shared JSON functionality across unit tests."""

from typing import List

from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import (
    common as java_common,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def generate(package: java_common.PackageIdentifier) -> List[java_common.JavaFile]:
    """
    Generate code for shared JSON functionality across unit tests.
    """
    blocks = [
        Stripped(
            f"""\
public static JsonNode readFromFile(Path path) {{
{I}final ObjectMapper objectMapper = new ObjectMapper();

{I}JsonNode node;
{I}try {{
{II}node = objectMapper.readTree(path.toFile());
{I}}} catch (JsonParseException exception) {{
{II}throw new IllegalStateException(
{III}"Expected the file to be a valid JSON, but it was not: "
{IIII}+ path + "; exception was: "
{IIII}+ exception);
{I}}} catch (JsonProcessingException exception) {{
{II}throw new IllegalStateException(
{III}"Expected the file to be a valid JSON, but it was not: "
{IIII}+ path + "; exception was: "
{IIII}+ exception);
{I}}} catch (IOException exception) {{
{II}throw new IllegalStateException(
{III}"Expected the file to be a valid JSON, but it was not: "
{IIII}+ path + "; exception was: "
{IIII}+ exception);
{I}}}
{I}return node;
}}"""
        ),
        Stripped(
            f"""\
/**
 *  Serialize something to a uniform JSON text
 *  such that we can use it for comparisons in the tests.
 */
public static JsonNode toJson(Object something)
{{
{I}if (something instanceof Boolean) {{
{II}return BooleanNode.valueOf((Boolean) something);
{I}}} else if (something instanceof Integer) {{
{II}return IntNode.valueOf((Integer) something);
{I}}} else if (something instanceof Double) {{
{II}return DoubleNode.valueOf((Double) something);
{I}}} else if (something instanceof String) {{
{II}return TextNode.valueOf((String) something);
{I}}} else if (something instanceof byte[]) {{
{II}return BinaryNode.valueOf((byte[]) something);
{I}}} else if (something instanceof IClass) {{
{II}return Jsonization.Serialize.toJsonObject((IClass) something);
{I}}} else {{
{II}throw new IllegalArgumentException(
{III}"The conversion of type "
{IIII}+ something.getClass().toString()
{IIII}+ " to a JSON node has not been defined: "
{IIII}+ something.toString());
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Infer the node kind of the JSON node.
 */
// TODO check whether this is really necessary
private static String getNodeKind(JsonNode node)
{{
{I}if (node.isArray()) {{
{II}return "array";
{I}}} else if (node.isObject()) {{
{II}return "object";
{I}}} else if (node.isValueNode()) {{
{II}return "value";
{I}}} else {{
{II}throw new IllegalStateException(
{III}"Unhandled JsonNode: " + node.getNodeType().toString());
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "CommonJson.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.jsonization.Jsonization;
import {package}.types.impl.*;
import {package}.types.model.*;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.*;
import java.io.IOException;
import java.nio.file.Path;

public final class CommonJson {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // public final class CommonJson

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Infer constraints representable in common schemas such as JSON Schema or XSD."""

from aas_core_codegen.infer_for_schema import (
    match,
    _len,
    _pattern,
    _inline,
    _stringify,
    _types,
)

LenConstraint = _types.LenConstraint
PatternConstraint = _types.PatternConstraint
SetOfPrimitivesConstraint = _types.SetOfPrimitivesConstraint
SetOfEnumerationLiteralsConstraint = _types.SetOfEnumerationLiteralsConstraint

Constraints = _types.Constraints
ConstraintsByValue = _types.ConstraintsByValue

infer_constraints_by_class = _inline.infer_constraints_by_class

tightening_steps_from_other_to_that_constraints = (
    _inline.tightening_steps_from_other_to_that_constraints
)

dump = _stringify.dump



"""Generate the functions related to filesystem used across the tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import common as golang_common
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the functions related to filesystem used across the tests."""
    blocks = [
        Stripped(
            """\
package aastesting"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"encoding/json"
{I}"fmt"
{I}"io/fs"
{I}"os"
{I}"path/filepath"
)"""
        ),
        Stripped(
            f"""\
// Read the content of `pth` and parse it as a JSON.
//
// If any errors, panic.
func MustReadJsonable(pth string) (jsonable interface{{}}) {{
{I}bb, err := os.ReadFile(pth)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to read the content of: %s",
{IIII}pth,
{III}),
{II})
{I}}}

{I}err = json.Unmarshal(bb, &jsonable)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to parse the content of %s as JSON: %s",
{IIII}pth, err.Error(),
{III}),
{II})
{I}}}
{I}return
}}"""
        ),
        Stripped(
            f"""\
func FindFilesBySuffixRecursively(root, suffix string) []string {{
{I}var a []string

{I}_, statErr := os.Stat(root)
{I}if statErr != nil {{
{II}if os.IsNotExist(statErr) {{
{III}return a
{II}}}

{II}panic(fmt.Sprintf("Failed to stat %s: %s", root, statErr.Error()))
{I}}}

{I}err := filepath.WalkDir(root, func(s string, d fs.DirEntry, e error) error {{
{II}if e != nil {{
{III}return e
{II}}}
{II}if filepath.Ext(d.Name()) == suffix {{
{III}a = append(a, s)
{II}}}
{II}return nil
{I}}})

{I}if err != nil {{
{II}panic(fmt.Sprintf("Failed to walk %s: %s", root, err.Error()))
{I}}}
{I}return a
}}"""
        ),
        golang_common.WARNING,
    ]  # type: List[Stripped]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the Java code for shared JSON functionality across unit tests."""

from typing import List

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import INDENT as I, INDENT2 as II


def generate(
    package: java_common.PackageIdentifier, symbol_table: intermediate.SymbolTable
) -> List[java_common.JavaFile]:
    """
    Generate the Java code for shared JSON functionality across unit tests.
    """
    blocks = []  # type: List[str]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
public static {cls_name_java} loadMaximal{cls_name_java}() throws IOException {{
{I}final Path path = Paths.get(
{II}Common.TEST_DATA_DIR,
{II}"Json",
{II}"Expected",
{II}{java_common.string_literal(cls_name_json)},
{II}"maximal.json");

{I}final JsonNode node = CommonJson.readFromFile(path);

{I}return Jsonization.Deserialize.deserialize{cls_name_java}(node);
}} // public static {cls_name_java} loadMaximal{cls_name_java}

public static {cls_name_java} loadMinimal{cls_name_java}() throws IOException {{
{I}final Path path = Paths.get(
{II}Common.TEST_DATA_DIR,
{II}"Json",
{II}"Expected",
{II}{java_common.string_literal(cls_name_json)},
{II}"minimal.json");

{I}final JsonNode node = CommonJson.readFromFile(path);

{I}return Jsonization.Deserialize.deserialize{cls_name_java}(node);
}} // public static {cls_name_java} loadMinimal{cls_name_java}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "CommonJsonization.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import {package}.jsonization.Jsonization;
import {package}.types.impl.*;
import com.fasterxml.jackson.databind.JsonNode;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;

/**
 * Provide methods to load instances from JSON test data.
 */
public final class CommonJsonization {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class CommonJsonization

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
